# Copyright 2019-2022 The ASReview Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

__all__ = [
    "DoubleBalance",
    "SimpleBalance",
    "UndersampleBalance",
    "get_balance_class",
    "get_balance_model",
    "list_balance_strategies",
]

from asreview.models.balance.double import DoubleBalance
from asreview.models.balance.simple import SimpleBalance
from asreview.models.balance.triple import TripleBalance
from asreview.models.balance.undersample import UndersampleBalance
from asreview.models.balance.utils import get_balance_class
from asreview.models.balance.utils import get_balance_model
from asreview.models.balance.utils import list_balance_strategies

"""Balance strategies to rebalance and reorder the training data.

There are several balance strategies that rebalance and reorder the
training data. This is sometimes necessary, because the data is often
very imbalanced: there are many more papers that should be excluded than
included (otherwise, automation cannot help much anyway).

There are several balance strategies available. In configuration
files, parameters are found under the section ``[balance_param]``.
"""



# Copyright 2019-2022 The ASReview Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

__all__ = ["LogisticClassifier"]

import logging

from sklearn.linear_model import LogisticRegression

from asreview.models.classifiers.base import BaseTrainClassifier
from asreview.models.classifiers.utils import _set_class_weight


class LogisticClassifier(BaseTrainClassifier):
    """
    Logistic regression classifier (``logistic``).

    The Logistic regressions classifier is an implementation based
    on the sklearn Logistic regressions classifier.

    Arguments
    ---------
    C: float
        Parameter inverse to the regularization strength of the model.
    class_weight: float
        Class weight of the inclusions.
    random_state: int, asreview.utils.SeededRandomState
        Random state for the model.
    n_jobs: int
        Number of CPU cores used.
    """

    name = "logistic"
    label = "Logistic regression"

    def __init__(self, C=1.0, class_weight=1.0, random_state=None, n_jobs=1):
        super(LogisticClassifier, self).__init__()
        self.C = C
        self.class_weight = class_weight
        self.n_jobs = n_jobs

        self._model = LogisticRegression(
            solver="liblinear",
            C=C,
            class_weight=_set_class_weight(class_weight),
            n_jobs=n_jobs,
            random_state=random_state,
        )
        logging.debug(self._model)

    def full_hyper_space(self):
        from hyperopt import hp

        hyper_choices = {}
        hyper_space = {
            "mdl_C": hp.lognormal("mdl_C", 0, 1),
            "mdl_class_weight": hp.lognormal("mdl_class_weight", 0, 1),
        }
        return hyper_space, hyper_choices



# Copyright 2019-2022 The ASReview Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

__all__ = ["LSTMBaseClassifier"]

import logging

try:
    import tensorflow as tf
    from tensorflow.keras import optimizers
    from tensorflow.keras.layers import LSTM
    from tensorflow.keras.layers import Dense
    from tensorflow.keras.layers import Embedding
    from tensorflow.keras.models import Sequential
except ImportError:
    TF_AVAILABLE = False
else:
    TF_AVAILABLE = True
    try:
        tf.logging.set_verbosity(tf.logging.ERROR)
    except AttributeError:
        logging.getLogger("tensorflow").setLevel(logging.ERROR)

import numpy as np
import scipy

from asreview.models.classifiers.base import BaseTrainClassifier
from asreview.models.classifiers.utils import _set_class_weight


def _check_tensorflow():
    if not TF_AVAILABLE:
        raise ImportError("Install tensorflow package to use" " LSTM-base.")


class LSTMBaseClassifier(BaseTrainClassifier):
    """LSTM-base classifier (``lstm-base``).

    LSTM model that consists of an embedding layer, LSTM layer with one
    output, dense layer, and a single sigmoid output node. Use the
    :class:`asreview.models.feature_extraction.EmbeddingLSTM` feature extraction
    method. Currently not so well optimized and slow.

    .. note::

        This model requires ``tensorflow`` to be installed. Use ``pip install
        asreview[tensorflow]`` or install all optional ASReview dependencies
        with ``pip install asreview[all]``

    Arguments
    ---------
    embedding_matrix: numpy.ndarray
        Embedding matrix to use with LSTM model.
    backwards: bool
        Whether to have a forward or backward LSTM.
    dropout: float
        Value in [0, 1.0) that gives the dropout and recurrent
        dropout rate for the LSTM model.
    optimizer: str
        Optimizer to use.
    lstm_out_width: int
        Output width of the LSTM.
    learn_rate: float
        Learn rate multiplier of default learning rate.
    dense_width: int
        Size of the dense layer of the model.
    verbose: int
        Verbosity.
    batch_size: int
        Size of the batch size for the LSTM model.
    epochs: int
        Number of epochs to train the LSTM model.
    shuffle: bool
        Whether to shuffle the data before starting to train.
    class_weight: float
        Class weight for the included papers.
    """

    name = "lstm-base"
    label = "LSTM classic"

    def __init__(
        self,
        embedding_matrix=None,
        backwards=True,
        dropout=0.4,
        optimizer="rmsprop",
        lstm_out_width=20,
        learn_rate=1.0,
        dense_width=128,
        verbose=0,
        batch_size=32,
        epochs=35,
        shuffle=False,
        class_weight=30.0,
    ):
        """Initialize the LSTM base model"""
        super(LSTMBaseClassifier, self).__init__()
        self.embedding_matrix = embedding_matrix
        self.backwards = backwards
        self.dropout = dropout
        self.optimizer = optimizer
        self.lstm_out_width = lstm_out_width
        self.learn_rate = learn_rate
        self.dense_width = dense_width
        self.verbose = verbose
        self.batch_size = batch_size
        self.epochs = epochs
        self.shuffle = shuffle
        self.class_weight = class_weight
        self._model = None
        self.sequence_length = None

    def fit(self, X, y):
        # check is tensorflow is available
        _check_tensorflow()

        if scipy.sparse.isspmatrix(X):
            X = X.toarray()

        sequence_length = X.shape[1]
        if self._model is None or sequence_length != self.sequence_length:
            self.sequence_length = sequence_length
            self._model = _create_lstm_base_model(
                embedding_matrix=self.embedding_matrix,
                backwards=self.backwards,
                dropout=self.dropout,
                optimizer=self.optimizer,
                max_sequence_length=self.sequence_length,
                lstm_out_width=self.lstm_out_width,
                dense_width=self.dense_width,
                learn_rate=self.learn_rate,
                verbose=self.verbose,
            )

        self._model.fit(
            X,
            y,
            batch_size=self.batch_size,
            epochs=self.epochs,
            shuffle=self.shuffle,
            class_weight=_set_class_weight(self.class_weight),
            verbose=self.verbose,
        )

    def predict_proba(self, X):
        if scipy.sparse.issparse(X):
            X = X.toarray()
        pos_pred = self._model.predict(X, verbose=self.verbose)
        neg_pred = 1 - pos_pred
        return np.hstack([neg_pred, pos_pred])

    def full_hyper_space(self):
        from hyperopt import hp

        hyper_choices = {}
        hyper_space = {
            "mdl_dropout": hp.uniform("mdl_dropout", 0, 0.9),
            "mdl_lstm_out_width": hp.quniform("mdl_lstm_out_width", 1, 50, 1),
            "mdl_dense_width": hp.quniform("mdl_dense_width", 1, 200, 1),
            "mdl_learn_rate_mult": hp.lognormal("mdl_learn_rate_mult", 0, 1),
        }
        return hyper_space, hyper_choices

    @property
    def default_param(self):
        defaults = super(LSTMBaseClassifier, self).default_param
        defaults.pop("embedding_matrix")
        return defaults


def _create_lstm_base_model(
    embedding_matrix,
    backwards=True,
    dropout=0.4,
    optimizer="rmsprop",
    max_sequence_length=1000,
    lstm_out_width=20,
    dense_width=128,
    learn_rate=1.0,
    verbose=1,
):
    """Return callable lstm model.
    Returns
    -------
    callable:
        A function that return the Keras Sklearn model when
        called.

    """

    # check is tensorflow is available
    _check_tensorflow()

    model = Sequential()

    # add first embedding layer with pretrained wikipedia weights
    model.add(
        Embedding(
            embedding_matrix.shape[0],
            embedding_matrix.shape[1],
            weights=[embedding_matrix],
            input_length=max_sequence_length,
            trainable=False,
        )
    )

    # add LSTM layer
    model.add(
        LSTM(
            lstm_out_width,
            input_shape=(max_sequence_length,),
            go_backwards=backwards,
            dropout=dropout,
            recurrent_dropout=dropout,
        )
    )

    # add Dense layer with relu activation
    model.add(
        Dense(
            dense_width,
            activation="relu",
        )
    )

    # add Dense layer
    model.add(Dense(1, activation="sigmoid"))

    optimizer_fn = _get_optimizer(optimizer, learn_rate)

    # Compile model
    model.compile(
        loss="binary_crossentropy", optimizer=optimizer_fn, metrics=["acc"]
    )

    if verbose >= 1:
        model.summary(verbose=verbose)

    return model


def _get_optimizer(optimizer, lr_mult=1.0):
    "Get optimizer with correct learning rate."
    if optimizer == "sgd":
        return optimizers.SGD(learning_rate=0.01 * lr_mult)
    elif optimizer == "rmsprop":
        return optimizers.RMSprop(learning_rate=0.001 * lr_mult)
    elif optimizer == "adagrad":
        return optimizers.Adagrad(learning_rate=0.01 * lr_mult)
    elif optimizer == "adam":
        return optimizers.Adam(learning_rate=0.001 * lr_mult)
    elif optimizer == "nadam":
        return optimizers.Nadam(learning_rate=0.002 * lr_mult)
    raise NotImplementedError



# Copyright 2019-2022 The ASReview Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

__all__ = ["PaperRecord"]

import os

import pandas as pd

from asreview.config import LABEL_NA
from asreview.utils import format_to_str


def preview_record(record, w_title=80, w_authors=40, automatic_width=False):
    """Return a single line preview string for record i.

    Arguments
    ---------
    record: PaperRecord
        The paperRecord to preview.
    w_title: int
        Width to be allocated for the title of the paper.
    w_authors: int
        Width to be allocated for the authors of the paper.
    automatic_width: bool
        If true, compute w_title, w_authors from the console width.

    Returns
    -------
    str:
        A string that previews a paper record.
    """
    if automatic_width:
        term_width = os.get_terminal_size().columns
        width_available = term_width - 7
        w_title = round((2 / 3) * width_available)
        w_authors = width_available - w_title
    title_str = ""
    author_str = ""
    heading = record.title
    if heading is None:
        heading = record.abstract
    if heading is not None:
        if len(heading) > w_title:
            title_str = heading[: w_title - 2] + ".."
        else:
            title_str = heading

    if record.authors is not None:
        cur_authors = format_to_str(record.authors)
        if len(cur_authors) > w_authors:
            author_str = cur_authors[: w_authors - 2] + ".."
        else:
            author_str = cur_authors
    format_str = "{0: <" + str(w_title) + "}   " + "{1: <" + str(w_authors)
    format_str += "}"
    prev_str = format_str.format(title_str, author_str)
    return prev_str


def format_record(record, use_cli_colors=True):
    """Format one record for displaying in the CLI.

    Arguments
    ---------
    record: PaperRecord
        The paperRecord to format.
    use_cli_colors: bool
        Some terminals support colors, set to True to use them.

    Returns
    -------
    str:
        A string including title, abstracts and authors.
    """
    if record.title is not None:
        title = record.title
        if use_cli_colors:
            title = "\033[95m" + title + "\033[0m"
        title += "\n"
    else:
        title = ""

    if record.authors is not None and len(record.authors) > 0:
        authors = format_to_str(record.authors) + "\n"
    else:
        authors = ""

    if record.abstract is not None and len(record.abstract) > 0:
        abstract = record.abstract
        abstract = "\n" + abstract + "\n"
    else:
        abstract = ""

    if record.included == 0:
        label = "IRRELEVANT"
    elif record.included == 1:
        label = "RELEVANT"
    else:
        label = ""

    header = f"---{record.record_id}---{label}---"

    return f"\n{header:-<60}\n{title}{authors}{abstract}"


class PaperRecord:
    """A single record from a paper in a systematic review.

    Arguments
    ---------
    record_id: int
        Some identifier for this record.
    title: str
        Paper title.
    abstract: str
        Paper abstract.
    authors: str, list
        Authors of the paper.
    notes: str, list
        Notes of the paper.
    keywords: str, list
        Keywords of the paper.
    included: int
        Current label of the paper. No label is indicated by
        asreview.config.LABEL_NA (== -1).
    kwargs: dict
        Any extra keyword arguments will be put in self.extra_fields.
    """

    def __init__(self, record_id, column_spec={}, **kwargs):
        for attr in [
            "title",
            "abstract",
            "authors",
            "notes",
            "keywords",
            "doi",
            "url",
            "included",
        ]:
            if attr in column_spec:
                col = column_spec[attr]
            elif attr in kwargs:
                col = attr
            else:
                col = None

            attr_val = kwargs.pop(col, None)
            if attr_val is not None and pd.isna(attr_val):
                attr_val = None
            setattr(self, attr, attr_val)

        self.record_id = record_id
        if self.included is None:
            self.included = LABEL_NA
        else:
            self.included = int(self.included)

        self.extra_fields = kwargs

        for attr, val in self.extra_fields.items():
            if not isinstance(val, list) and pd.isna(val):
                self.extra_fields[attr] = None

    def __str__(self):
        return format_record(self)

    @property
    def text(self):
        """Create a single string from title + abstract.

        Returns
        -------
        str:
            Concatenated string from title + abstract.
        """
        title = self.title
        abstract = self.abstract
        if title is None:
            title = ""
        if abstract is None:
            abstract = ""
        return title + " " + abstract

    @property
    def heading(self):
        """Return the title of the paper."""
        if self.title is None:
            return ""
        return self.title

    @property
    def body(self):
        """Return the abstract of the paper."""
        if self.abstract is None:
            return ""
        return self.abstract



# Copyright 2019-2022 The ASReview Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

__all__ = ["ASReviewData", "load_data"]

import hashlib
from io import StringIO
from pathlib import Path

import numpy as np
import pandas as pd
from pandas.api.types import is_object_dtype
from pandas.api.types import is_string_dtype

from asreview.config import COLUMN_DEFINITIONS
from asreview.config import LABEL_NA
from asreview.datasets import DatasetManager
from asreview.datasets import DatasetNotFoundError
from asreview.exceptions import BadFileFormatError
from asreview.io import PaperRecord
from asreview.io.utils import convert_keywords
from asreview.io.utils import type_from_column
from asreview.utils import _entry_points
from asreview.utils import _get_filename_from_url
from asreview.utils import is_iterable
from asreview.utils import is_url


def load_data(name, **kwargs):
    """Load data from file, URL, or plugin.

    Parameters
    ----------
    name: str, pathlib.Path
        File path, URL, or alias of extension dataset.
    **kwargs:
        Keyword arguments passed to the reader.

    Returns
    -------
    asreview.ASReviewData:
        Inititalized ASReview data object.
    """

    # check is file or URL
    if is_url(name) or Path(name).exists():
        return ASReviewData.from_file(name, **kwargs)

    # check if dataset is plugin dataset
    try:
        return ASReviewData.from_extension(name, **kwargs)
    except DatasetNotFoundError:
        pass

    # Could not find dataset, return None.
    raise FileNotFoundError(f"File, URL, or dataset does not exist: '{name}'")


class ASReviewData:
    """Data object to the dataset with texts, labels, DOIs etc.

    Arguments
    ---------
    df: pandas.DataFrame
        Dataframe containing the data for the ASReview data object.
    column_spec: dict
        Specification for which column corresponds to which standard
        specification. Key is the standard specification, key is which column
        it is actually in. Default: None.

    Attributes
    ----------
    record_ids: numpy.ndarray
        Return an array representing the data in the Index.
    texts: numpy.ndarray
        Returns an array with either headings, bodies, or both.
    headings: numpy.ndarray
        Returns an array with dataset headings.
    title: numpy.ndarray
        Identical to headings.
    bodies: numpy.ndarray
        Returns an array with dataset bodies.
    abstract: numpy.ndarray
        Identical to bodies.
    notes: numpy.ndarray
        Returns an array with dataset notes.
    keywords: numpy.ndarray
        Returns an array with dataset keywords.
    authors: numpy.ndarray
        Returns an array with dataset authors.
    doi: numpy.ndarray
        Returns an array with dataset DOI.
    included: numpy.ndarray
        Returns an array with document inclusion markers.
    final_included: numpy.ndarray
        Pending deprecation! Returns an array with document inclusion markers.
    labels: numpy.ndarray
        Identical to included.

    """

    def __init__(self, df=None, column_spec=None):
        self.df = df
        self.prior_idx = np.array([], dtype=int)
        self.max_idx = max(df.index.values) + 1

        # Infer column specifications if it is not given.
        if column_spec is None:
            self.column_spec = {}
            for col_name in list(df):
                data_type = type_from_column(col_name, COLUMN_DEFINITIONS)
                if data_type is not None:
                    self.column_spec[data_type] = col_name
        else:
            self.column_spec = column_spec

        if "included" not in self.column_spec:
            self.column_spec["included"] = "included"

    def __len__(self):
        if self.df is None:
            return 0
        return len(self.df.index)

    def hash(self):
        """Compute a hash from the dataset.

        Returns
        -------
        str:
            SHA1 hash, computed from the titles/abstracts of the dataframe.
        """
        if (
            len(self.df.index) < 1000 and self.bodies is not None
        ) or self.texts is None:
            texts = " ".join(self.bodies)
        else:
            texts = " ".join(self.texts)
        return hashlib.sha1(
            " ".join(texts).encode(encoding="UTF-8", errors="ignore")
        ).hexdigest()

    @classmethod
    def from_file(cls, fp, reader=None):
        """Create instance from supported file format.

        It works in two ways; either manual control where the conversion
        functions are supplied or automatic, where it searches in the entry
        points for the right conversion functions.

        Arguments
        ---------
        fp: str, pathlib.Path
            Read the data from this file or url.
        reader: class
            Reader to import the file.
        """

        if reader is not None:
            return cls(reader.read_data(fp))

        # get the filename from a url else file path
        if is_url(fp):
            fn = _get_filename_from_url(fp)
        else:
            fn = Path(fp).name

        try:
            reader = _entry_points(
                group="asreview.readers")[Path(fn).suffix].load()
        except Exception:
            raise BadFileFormatError(f"Importing file {fp} not possible.")

        df, column_spec = reader.read_data(fp)

        return cls(df, column_spec=column_spec)

    @classmethod
    def from_extension(cls, name, reader=None):
        """Load a dataset from extension.

        Arguments
        ---------
        fp: str, pathlib.Path
            Read the data from this file or url.
        reader: class
            Reader to import the file.
        """

        dataset = DatasetManager().find(name)

        if dataset.filepath:
            fp = dataset.filepath
        else:
            # build dataset to temporary file
            reader = dataset.reader()
            fp = StringIO(dataset.to_file())

        if reader is None:
            # get the filename from a url else file path
            if is_url(fp):
                fn = _get_filename_from_url(fp)
            else:
                fn = Path(fp).name

            try:
                reader = _entry_points(
                    group="asreview.readers")[Path(fn).suffix].load()
            except Exception:
                raise BadFileFormatError(f"Importing file {fp} not possible.")

        df, column_spec = reader.read_data(fp)

        return cls(df, column_spec=column_spec)

    def record(self, i, by_index=True):
        """Create a record from an index.

        Arguments
        ---------
        i: int, iterable
            Index of the record, or list of indices.
        by_index: bool
            If True, take the i-th value as used internally by the review.
            If False, take the record with record_id==i.

        Returns
        -------
        PaperRecord
            The corresponding record if i was an integer, or a list of records
            if i was an iterable.
        """
        if not is_iterable(i):
            index_list = [i]
        else:
            index_list = i

        if by_index:
            records = [
                PaperRecord(
                    **self.df.iloc[j],
                    column_spec=self.column_spec,
                    record_id=self.df.index.values[j],
                )
                for j in index_list
            ]
        else:
            records = [
                PaperRecord(
                    **self.df.loc[j, :], record_id=j, column_spec=self.column_spec
                )
                for j in index_list
            ]

        if is_iterable(i):
            return records
        return records[0]

    @property
    def record_ids(self):
        return self.df.index.values

    @property
    def texts(self):
        if self.title is None:
            return self.abstract
        if self.abstract is None:
            return self.title

        cur_texts = np.array(
            [self.title[i] + " " + self.abstract[i] for i in range(len(self))],
            dtype=object,
        )
        return cur_texts

    @property
    def headings(self):
        return self.title

    @property
    def title(self):
        try:
            return self.df[self.column_spec["title"]].values
        except KeyError:
            return None

    @property
    def bodies(self):
        return self.abstract

    @property
    def abstract(self):
        try:
            return self.df[self.column_spec["abstract"]].values
        except KeyError:
            return None

    @property
    def notes(self):
        try:
            return self.df[self.column_spec["notes"]].values
        except KeyError:
            return None

    @property
    def keywords(self):
        try:
            return self.df[self.column_spec["keywords"]].apply(convert_keywords).values
        except KeyError:
            return None

    @property
    def authors(self):
        try:
            return self.df[self.column_spec["authors"]].values
        except KeyError:
            return None

    @property
    def doi(self):
        try:
            return self.df[self.column_spec["doi"]].values
        except KeyError:
            return None

    @property
    def url(self):
        try:
            return self.df[self.column_spec["url"]].values
        except KeyError:
            return None

    def get(self, name):
        "Get column with name."
        try:
            return self.df[self.column_spec[name]].values
        except KeyError:
            return self.df[name].values

    @property
    def prior_data_idx(self):
        "Get prior_included, prior_excluded from dataset."
        convert_array = np.full(self.max_idx, 999999999)
        convert_array[self.df.index.values] = np.arange(len(self.df.index))
        return convert_array[self.prior_idx]

    @property
    def included(self):
        return self.labels

    @included.setter
    def included(self, labels):
        self.labels = labels

    @property  # pending deprecation
    def final_included(self):
        return self.labels

    @final_included.setter  # pending deprecation
    def final_included(self, labels):
        self.labels = labels

    @property
    def labels(self):
        try:
            column = self.column_spec["included"]
            return self.df[column].values
        except KeyError:
            return None

    @labels.setter
    def labels(self, labels):
        try:
            column = self.column_spec["included"]
            self.df[column] = labels
        except KeyError:
            self.df["included"] = labels

    def prior_labels(self, state, by_index=True):
        """Get the labels that are marked as 'prior'.

        state: BaseState
            Open state that contains the label information.
        by_index: bool
            If True, return internal indexing.
            If False, return record_ids for indexing.

        Returns
        -------
        numpy.ndarray
            Array of indices that have the 'prior' property.
        """
        prior_indices = state.get_priors()["record_id"].to_list()

        if by_index:
            return np.array(prior_indices, dtype=int)
        else:
            return self.df.index.values[prior_indices]

    def to_file(
            self, fp, labels=None, ranking=None, writer=None, keep_old_labels=False):
        """Export data object to file.

        RIS, CSV, TSV and Excel are supported file formats at the moment.

        Arguments
        ---------
        fp: str
            Filepath to export to.
        labels: list, numpy.ndarray
            Labels to be inserted into the dataframe before export.
        ranking: list, numpy.ndarray
            Optionally, dataframe rows can be reordered.
        writer: class
            Writer to export the file.
        keep_old_labels: bool
            If True, the old labels are kept in a column 'asreview_label_to_validate'.
            Default False.
        """
        df = self.to_dataframe(
            labels=labels, ranking=ranking, keep_old_labels=keep_old_labels
        )

        if writer is not None:
            writer.write_data(df, fp, labels=labels, ranking=ranking)
        else:
            best_suffix = None

            for entry in _entry_points(group="asreview.writers"):
                if Path(fp).suffix == entry.name:
                    if best_suffix is None or len(entry.name) > len(best_suffix):
                        best_suffix = entry.name

            if best_suffix is None:
                raise BadFileFormatError(
                    f"Error exporting file {fp}, no capabilities "
                    "for exporting such a file."
                )

            writer = _entry_points(group="asreview.writers")[best_suffix].load()
            writer.write_data(df, fp, labels=labels, ranking=ranking)

    def to_dataframe(self, labels=None, ranking=None, keep_old_labels=False):
        """Create new dataframe with updated label (order).

        Arguments
        ---------
        labels: list, numpy.ndarray
            Current labels will be overwritten by these labels
            (including unlabelled). No effect if labels is None.
        ranking: list
            Reorder the dataframe according to these record_ids.
            Default ordering if ranking is None.
        keep_old_labels: bool
            If True, the old labels are kept in a column 'asreview_label_to_validate'.
            Default False.

        Returns
        -------
        pandas.DataFrame
            Dataframe of all available record data.
        """
        result_df = pd.DataFrame.copy(self.df)
        col_label = self.column_spec["included"]

        # if there are labels, add them to the frame
        if labels is not None:
            # unnest list of nested (record_id, label) tuples
            labeled_record_ids = [x[0] for x in labels]
            labeled_values = [x[1] for x in labels]

            if keep_old_labels:
                result_df["asreview_label_to_validate"] = \
                    result_df[col_label].replace(LABEL_NA, None).astype("Int64")

            # remove the old results and write the values
            result_df[col_label] = LABEL_NA
            result_df.loc[labeled_record_ids, col_label] = labeled_values
            result_df[col_label] = result_df[col_label] \
                .replace(LABEL_NA, None).astype("Int64")

        # if there is a ranking, apply this ranking as order
        if ranking is not None:
            # sort the datasets based on the ranking
            result_df = result_df.loc[ranking]
            # append a column with 1 to n
            result_df["asreview_ranking"] = np.arange(1, len(result_df) + 1)

        return result_df

    def duplicated(self, pid="doi"):
        """Return boolean Series denoting duplicate rows.

        Identify duplicates based on titles and abstracts and if available,
        on a persistent identifier (PID) such as the Digital Object Identifier
        (`DOI <https://www.doi.org/>`_).

        Arguments
        ---------
        pid: string
            Which persistent identifier to use for deduplication.
            Default is 'doi'.

        Returns
        -------
        pandas.Series
            Boolean series for each duplicated rows.
        """
        if pid in self.df.columns:
            # in case of strings, strip whitespaces and replace empty strings with None
            if is_string_dtype(self.df[pid]) or is_object_dtype(self.df[pid]):
                s_pid = self.df[pid].str.strip().replace("", None)
                if pid == "doi":
                    s_pid = s_pid.str.lower().str.replace(
                        r"^https?://(www\.)?doi\.org/", "", regex=True
                    )
            else:
                s_pid = self.df[pid]

            # save boolean series for duplicates based on persistent identifiers
            s_dups_pid = (s_pid.duplicated()) & (s_pid.notnull())
        else:
            s_dups_pid = None

        # get the texts, clean them and replace empty strings with None
        s = (
            pd.Series(self.texts)
            .str.replace("[^A-Za-z0-9]", "", regex=True)
            .str.lower()
            .str.strip()
            .replace("", None)
        )

        # save boolean series for duplicates based on titles/abstracts
        s_dups_text = (s.duplicated()) & (s.notnull())

        # final boolean series for all duplicates
        if s_dups_pid is not None:
            s_dups = s_dups_pid | s_dups_text
        else:
            s_dups = s_dups_text

        return s_dups

    def drop_duplicates(self, pid="doi", inplace=False, reset_index=True):
        """Drop duplicate records.

        Drop duplicates based on titles and abstracts and if available,
        on a persistent identifier (PID) such the Digital Object Identifier
        (`DOI <https://www.doi.org/>`_).

        Arguments
        ---------
        pid: string, default 'doi'
            Which persistent identifier to use for deduplication.
        inplace: boolean, default False
            Whether to modify the DataFrame rather than creating a new one.
        reset_index: boolean, default True
            If True, the existing index column is reset to the default integer index.

        Returns
        -------
        pandas.DataFrame or None
            DataFrame with duplicates removed or None if inplace=True
        """
        df = self.df[~self.duplicated(pid)]

        if reset_index:
            df = df.reset_index(drop=True)
        if inplace:
            self.df = df
            return
        return df



# Copyright 2019-2022 The ASReview Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

__all__ = [
    "Doc2Vec",
    "EmbeddingIdf",
    "EmbeddingLSTM",
    "SBERT",
    "Tfidf",
    "get_feature_class",
    "get_feature_model",
    "list_feature_extraction",
]

from asreview.models.feature_extraction.doc2vec import Doc2Vec
from asreview.models.feature_extraction.embedding_idf import EmbeddingIdf
from asreview.models.feature_extraction.embedding_lstm import EmbeddingLSTM
from asreview.models.feature_extraction.sbert import SBERT
from asreview.models.feature_extraction.tfidf import Tfidf
from asreview.models.feature_extraction.utils import get_feature_class
from asreview.models.feature_extraction.utils import get_feature_model
from asreview.models.feature_extraction.utils import list_feature_extraction

"""Feature extraction converts texts into features.

Feature extraction is the process of converting a list of texts into some kind
of feature matrix.

There are several feature extraction algorithms available. In configuration
files, parameters are found under the section ``[feature_param]``.

"""



import random
import time
from typing import Union

from flask.testing import FlaskClient

import asreview.webapp.tests.utils.crud as crud
import asreview.webapp.tests.utils.misc as misc
from asreview.project import ASReviewProject
from asreview.webapp.authentication.models import Project
from asreview.webapp.tests.utils.config_parser import all_users
from asreview.webapp.tests.utils.config_parser import get_user
from asreview.webapp.tests.utils.misc import get_project_id


def process_response(response):
    """Breaks response in a tuple containing the status code
    and json data."""
    return (response.status_code, response.json)


# ########################
# General API calls
# ########################


def call_root_url(client):
    response = client.get("/")
    status_code, data = process_response(response)
    return (status_code, data, response.text)


def call_boot_url(client):
    response = client.get("/boot")
    return process_response(response)


# ########################
# Authentication API calls
# ########################


def signin_user(client, user):
    """Signs in a user through the api"""
    # If a password is not set, we need to get it
    if not hasattr(user, "password"):
        users = all_users()
        user.password = users[user.identifier].password
    # request
    response = client.post(
        "/auth/signin", data={"email": user.identifier, "password": user.password}
    )
    return process_response(response)


def signup_user(client, user):
    """Signs up a user through the api"""
    response = client.post(
        "/auth/signup",
        data={
            "identifier": user.email,
            "email": user.email,
            "name": user.name,
            "password": user.password,
            "affiliation": user.affiliation,
            "origin": "asreview",
        },
    )
    return process_response(response)


def signout_user(client):
    """Sign out user"""
    response = client.delete("/auth/signout")
    return process_response(response)


def confirm_user(client, user):
    response = client.post(
        "/auth/confirm_account", data={"user_id": user.id, "token": user.token}
    )
    return process_response(response)


def forgot_password(client, user):
    response = client.post("/auth/forgot_password", data={"email": user.email})
    return process_response(response)


def reset_password(client, user):
    response = client.post(
        "/auth/reset_password",
        data={"password": user.password, "token": user.token, "user_id": user.id},
    )
    return process_response(response)


def update_user(client, data):
    response = client.post("/auth/update_profile", data=data)
    return process_response(response)


def refresh(client):
    response = client.get("/auth/refresh")
    return process_response(response)


def get_profile(client: FlaskClient):
    response = client.get("/auth/get_profile")
    return process_response(response)


# ########################
# Teams API calls
# ########################


def invite(client, project, user):
    url = f"/api/invitations/projects/{get_project_id(project)}/users/{user.id}"
    response = client.post(url)
    return process_response(response)


def list_invitations(client):
    response = client.get("/api/invitations")
    return process_response(response)


def list_collaborators(client, project):
    response = client.get(f"/api/projects/{get_project_id(project)}/users")
    return process_response(response)


def accept_invitation(client, project):
    response = client.post(
        f"/api/invitations/projects/{get_project_id(project)}/accept", data={}
    )
    return process_response(response)


def reject_invitation(client, project):
    response = client.delete(
        f"/api/invitations/projects/{get_project_id(project)}/reject", data={}
    )
    return process_response(response)


def delete_invitation(client, project, user):
    response = client.delete(
        f"/api/invitations/projects/{get_project_id(project)}/users/{user.id}", data={}
    )
    return process_response(response)


def delete_collaboration(client, project, user):
    response = client.delete(
        f"/api/projects/{get_project_id(project)}/users/{user.id}", data={}
    )
    return process_response(response)


# ########################
# Project API calls
# ########################


def get_all_projects(client: FlaskClient):
    response = client.get("/api/projects")
    return process_response(response)


def create_project(
    client: FlaskClient,
    project_name: str,
    mode: str = "explore",
    authors: str = "authors",
    description: str = "description",
):
    response = client.post(
        "/api/projects/info",
        data={
            "mode": mode,
            "name": project_name,
            "authors": authors,
            "description": description,
        },
    )
    return process_response(response)


def create_project_from_dict(client: FlaskClient, data: dict):
    response = client.post(
        "/api/projects/info",
        data=data,
    )
    return process_response(response)


def update_project(
    client: FlaskClient,
    project: Union[Project, ASReviewProject],
    name: str = "name",
    mode: str = "explore",
    authors: str = "authors",
    description: str = "description",
):
    response = client.put(
        f"/api/projects/{get_project_id(project)}/info",
        data={
            "mode": mode,
            "name": name,
            "authors": authors,
            "description": description,
        },
    )
    return process_response(response)


def upgrade_project(client: FlaskClient, project: Union[Project, ASReviewProject]):
    response = client.get(f"/api/projects/{get_project_id(project)}/upgrade_if_old")
    return process_response(response)


def import_project(client: FlaskClient, asreview_file):
    response = client.post(
        "/api/projects/import_project",
        data={"file": (open(asreview_file, "rb"), "project.asreview")},
    )
    return process_response(response)


def get_project_stats(client: FlaskClient):
    response = client.get("/api/projects/stats")
    return process_response(response)


def get_demo_data(client: FlaskClient, subset: str):
    response = client.get(f"/api/datasets?subset={subset}")
    return process_response(response)


def upload_data_to_project(
    client: FlaskClient, project: Union[Project, ASReviewProject], data: dict
):
    response = client.post(
        f"/api/projects/{get_project_id(project)}/data",
        data=data,
    )
    return process_response(response)


def get_project_data(client: FlaskClient, project: Union[Project, ASReviewProject]):
    response = client.get(f"/api/projects/{get_project_id(project)}/data")
    return process_response(response)


def get_project_dataset_writer(
    client: FlaskClient, project: Union[Project, ASReviewProject]
):
    response = client.get(f"/api/projects/{get_project_id(project)}/dataset_writer")
    return process_response(response)


def search_project_data(
    client: FlaskClient, project: Union[Project, ASReviewProject], query: str
):
    response = client.get(f"/api/projects/{get_project_id(project)}/search?q={query}")
    return process_response(response)


def get_prior_random_project_data(
    client: FlaskClient, project: Union[Project, ASReviewProject]
):
    response = client.get(f"/api/projects/{get_project_id(project)}/prior_random")
    return process_response(response)


def label_random_project_data_record(
    client: FlaskClient, project: Union[Project, ASReviewProject], label: int
):
    # get random data
    _, data = get_prior_random_project_data(client, project)
    # select a specific record
    record = random.choice(data["result"])
    doc_id = record["id"]
    return label_project_record(client, project, doc_id, label, note="")


def label_project_record(
    client: FlaskClient,
    project: Union[Project, ASReviewProject],
    doc_id: int,
    label: str,
    prior: int = 1,
    note: str = "",
):
    response = client.post(
        f"/api/projects/{get_project_id(project)}/record/{doc_id}",
        data={"doc_id": doc_id, "label": label, "is_prior": prior, "note": note},
    )
    return process_response(response)


def update_label_project_record(
    client: FlaskClient,
    project: Union[Project, ASReviewProject],
    doc_id: int,
    label: str,
    prior: int = 1,
    note: str = "",
):
    response = client.put(
        f"/api/projects/{get_project_id(project)}/record/{doc_id}",
        data={"doc_id": doc_id, "label": label, "is_prior": prior, "note": note},
    )
    return process_response(response)


def get_labeled_project_data(
    client: FlaskClient, project: Union[Project, ASReviewProject]
):
    response = client.get(f"/api/projects/{get_project_id(project)}/labeled")
    return process_response(response)


def get_labeled_project_data_stats(
    client: FlaskClient, project: Union[Project, ASReviewProject]
):
    response = client.get(f"/api/projects/{get_project_id(project)}/labeled_stats")
    return process_response(response)


def get_project_algorithms_options(client: FlaskClient):
    response = client.get("/api/algorithms")
    return process_response(response)


def set_project_algorithms(
    client: FlaskClient, project: Union[Project, ASReviewProject], data: dict
):
    response = client.post(
        f"/api/projects/{get_project_id(project)}/algorithms", data=data
    )
    return process_response(response)


def get_project_algorithms(
    client: FlaskClient, project: Union[Project, ASReviewProject]
):
    response = client.get(f"/api/projects/{get_project_id(project)}/algorithms")
    return process_response(response)


def start_project_algorithms(
    client: FlaskClient, project: Union[Project, ASReviewProject]
):
    response = client.post(f"/api/projects/{get_project_id(project)}/start")
    return process_response(response)


def get_project_status(client: FlaskClient, project: Union[Project, ASReviewProject]):
    response = client.get(f"/api/projects/{get_project_id(project)}/status")
    return process_response(response)


def set_project_status(
    client: FlaskClient, project: Union[Project, ASReviewProject], status: str
):
    response = client.put(
        f"/api/projects/{get_project_id(project)}/status", data={"status": status}
    )
    return process_response(response)


def export_project_dataset(
    client: FlaskClient, project: Union[Project, ASReviewProject], format: str
):
    id = get_project_id(project)
    response = client.get(f"/api/projects/{id}/export_dataset?file_format={format}")
    return process_response(response)


def export_project(
    client: FlaskClient,
    project: Union[Project, ASReviewProject],
):
    response = client.get(f"/api/projects/{get_project_id(project)}/export_project")
    return process_response(response)


def get_project_progress(
    client: FlaskClient,
    project: Union[Project, ASReviewProject],
):
    response = client.get(f"/api/projects/{get_project_id(project)}/progress")
    return process_response(response)


def get_project_progress_density(
    client: FlaskClient,
    project: Union[Project, ASReviewProject],
):
    response = client.get(f"/api/projects/{get_project_id(project)}/progress_density")
    return process_response(response)


def get_project_progress_recall(
    client: FlaskClient,
    project: Union[Project, ASReviewProject],
):
    response = client.get(f"/api/projects/{get_project_id(project)}/progress_recall")
    return process_response(response)


def get_project_current_document(
    client: FlaskClient,
    project: Union[Project, ASReviewProject],
):
    response = client.get(f"/api/projects/{get_project_id(project)}/get_document")
    return process_response(response)


def delete_project(
    client: FlaskClient,
    project: Union[Project, ASReviewProject],
):
    response = client.delete(f"/api/projects/{get_project_id(project)}/delete")
    return process_response(response)


# ########################
# General procedures
# ########################


def create_and_signin_user(client, test_user_id=1):
    """Creates a user account and signs in with that account."""
    # signup user
    user = get_user(test_user_id)
    signup_user(client, user)
    # refresh user
    stored_user = crud.get_user_by_identifier(user.identifier)
    # signin user
    signin_user(client, user)
    # return the user
    return stored_user


def upload_label_set_and_start_model(client, project, dataset):
    """Uploads a dataset to a created project and adds and starts
    a random model."""
    # upload dataset
    upload_data_to_project(client, project, data=dataset)
    # label 2 random records
    label_random_project_data_record(client, project, 1)
    label_random_project_data_record(client, project, 0)
    # select a model
    model_data = misc.choose_project_algorithms()
    set_project_algorithms(client, project, data=model_data)
    # start the model
    start_project_algorithms(client, project)
    # make sure model is done
    time.sleep(10)



import urllib
from pathlib import Path

import pandas as pd
from pytest import mark

import asreview
from asreview.data.base import ASReviewData
from asreview.data.statistics import n_duplicates
from asreview.datasets import DatasetManager
from asreview.search import fuzzy_find


def exists(url):
    return urllib.request.urlopen(url).getcode() == 200


@mark.parametrize(
    "keywords,paper_id",
    [
        ("bronchogenic duplication cyst", 0),
        ("diagnositc accuracy microscopy female priority", 1),
        ("immunophenotiping", 4),
        ("Foregut report embryoogenesis", 4),
        ("Liu Adler", 0),
        ("Khoury cysts", 4),
        ("Isolated Edwards", 5),
        ("Kwintanilla-djeck Neck", 3),
        ("Cancer case computer contrast pancreatomy Yamada", 2),
    ],
)
def test_fuzzy_finder(keywords, paper_id):
    fp = Path("tests", "demo_data", "embase.csv")
    as_data = asreview.ASReviewData.from_file(fp)

    assert fuzzy_find(as_data, keywords)[0] == paper_id


@mark.parametrize(
    "data_name",
    [
        # datasets from the Van de Schoot et al. paper
        # https://github.com/asreview/paper-asreview/blob/master/index_v1.json
        "benchmark-nature:van_de_Schoot_2017",
        "benchmark-nature:Hall_2012",
        "benchmark-nature:Cohen_2006_ACEInhibitors",
        "benchmark-nature:Kwok_2020",
    ],
)
def test_datasets(data_name):
    data = DatasetManager().find(data_name)
    assert exists(data.filepath)


def test_duplicate_count():
    d = ASReviewData.from_file(Path("tests", "demo_data", "duplicate_records.csv"))

    assert n_duplicates(d) == 2


def test_deduplication():
    d_dups = ASReviewData.from_file(Path("tests", "demo_data", "duplicate_records.csv"))

    s_dups_bool = pd.Series(
        [
            False,
            True,
            False,
            True,
            False,
            False,
            False,
            False,
            False,
            False,
            False,
            False,
            False,
            False,
        ]
    )

    # test whether .duplicated() provides correct boolean series for duplicates
    pd.testing.assert_series_equal(d_dups.duplicated(), s_dups_bool, check_index=False)

    d_nodups = ASReviewData(
        pd.DataFrame(
            {
                "title": ["a", "b", "d", "e", "f", "g", "h", "i", "", "", "   ", "   "],
                "abstract": [
                    "lorem",
                    "lorem",
                    "lorem",
                    "lorem",
                    "lorem",
                    "lorem",
                    "lorem",
                    "lorem",
                    "",
                    "",
                    "   ",
                    "   ",
                ],
                "doi": [
                    "10.1",
                    "10.3",
                    None,
                    None,
                    "   ",
                    "   ",
                    None,
                    None,
                    "10.4",
                    "10.5",
                    "10.6",
                    "10.7",
                ],
                "some_column": [
                    "lorem",
                    "lorem",
                    "lorem",
                    "lorem",
                    "lorem",
                    "lorem",
                    "lorem",
                    "lorem",
                    "lorem",
                    "lorem",
                    "lorem",
                    "lorem",
                ],
            }
        )
    )

    # test whether .drop_duplicates() drops the duplicated records correctly
    pd.testing.assert_frame_equal(d_dups.drop_duplicates(), d_nodups.df)


def test_duplicated():
    # Create an instance of ASReviewData
    instance = ASReviewData(pd.DataFrame({
        'doi': ['https://www.doi.org/10.1000/xyz',
                'https://www.doi.org/10.1000/abc',
                'https://www.doi.org/10.1000/xyz',
                'https://www.doi.org/10.1000/XYZ',
                '10.1000/xyz',
                '10.1000/xyz',
                'http://www.doi.org/10.1000/xyz',
                'https://doi.org/10.1000/xyz'],
        'title': ['T1', 'T2', 'T3', 'T3', 'T4', 'T1', 'T2', 'T3'],
        'abstract': ['A1', 'A2', 'A3', 'A3', 'A4', 'A1', 'A2', 'A3']
    }))

    # Call the function and get the result
    result = instance.duplicated()

    # Check the result
    assert result.equals(pd.Series([False, False, True, True, True, True, True, True]))



# Copyright 2019-2022 The ASReview Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from asreview.data.base import ASReviewData
from asreview.data.base import load_data
from asreview.io.utils import list_readers
from asreview.io.utils import list_writers
from asreview.project import ASReviewProject
from asreview.project import open_state
from asreview.utils import asreview_path
from asreview.utils import get_data_home

from ._version import get_versions

__version__ = get_versions()["version"]
del get_versions

__all__ = [
    "asreview_path",
    "ASReviewData",
    "ASReviewProject",
    "get_data_home",
    "list_readers",
    "list_writers",
    "open_state",
]

# deprecated in __init__.py, use asreview.models.feature_extraction instead
from asreview._deprecated import _deprecated_func
from asreview.models.feature_extraction.embedding_lstm import load_embedding as _load_embedding    # NOQA
from asreview.models.feature_extraction.embedding_lstm import sample_embedding as _sample_embedding    # NOQA
from asreview.models.feature_extraction.embedding_lstm import text_to_features as _text_to_features    # NOQA


@_deprecated_func(
    "Importing load_embedding from asreview.load_embedding is deprecated, "
    "use asreview.models.feature_extraction.load_embedding instead"
)
def load_embedding(*args, **kwargs):
    return _load_embedding(*args, **kwargs)


@_deprecated_func(
    "Importing sample_embedding from asreview.sample_embedding is deprecated, "
    "use asreview.models.feature_extraction.sample_embedding instead"
)
def sample_embedding(*args, **kwargs):
    return _sample_embedding(*args, **kwargs)


@_deprecated_func(
    "Importing text_to_features from asreview.text_to_features is deprecated, "
    "use asreview.models.feature_extraction.text_to_features instead"
)
def text_to_features(*args, **kwargs):
    return _text_to_features(*args, **kwargs)



# Copyright 2019-2022 The ASReview Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


import datetime as dt
from pathlib import Path

from flask import Blueprint
from flask import current_app
from flask import jsonify
from flask import render_template_string
from flask import request
from flask_login import current_user
from flask_login import login_required
from flask_login import login_user
from flask_login import logout_user
from flask_mail import Mail
from flask_mail import Message
from sqlalchemy import and_
from sqlalchemy import or_
from sqlalchemy.exc import IntegrityError
from sqlalchemy.exc import SQLAlchemyError

from asreview.webapp import DB
from asreview.webapp.authentication.models import User
from asreview.webapp.authentication.oauth_handler import OAuthHandler

bp = Blueprint("auth", __name__, url_prefix="/auth")


def _has_email_configuration(app):
    return all([
        app.config.get("MAIL_SERVER", False),
        app.config.get("MAIL_USERNAME", False),
        app.config.get("MAIL_PASSWORD", False)
    ])


def perform_login_user(user):
    """Helper function to login a user"""
    return login_user(user, remember=True, duration=dt.timedelta(days=31))


# TODO: not sure if this file is the right place for this function
def send_forgot_password_email(user, request, cur_app):
    # do not send email in test environment
    if not cur_app.testing:
        # get necessary information out of user object
        name = user.name or "ASReview user"
        # get url of front-end
        root_url = request.headers.get("Origin")
        # create url that will be used in the email
        url = f"{root_url}/reset_password?user_id={user.id}&token={user.token}"
        # create a mailer
        mailer = Mail(cur_app)
        # open templates as string and render
        root_path = Path(cur_app.root_path)
        with open(
            root_path / "templates" / "emails" / "forgot_password.html", "r"
        ) as f:
            html_text = render_template_string(f.read(), name=name, url=url)
        with open(root_path / "templates" / "emails" / "forgot_password.txt", "r") as f:
            txt_text = render_template_string(f.read(), name=name, url=url)
        # create message
        msg = Message(
            "ASReview: forgot password",
            recipients=[user.email]
        )
        msg.body = txt_text
        msg.html = html_text
        return mailer.send(msg)


# TODO: not sure if this file is the right place for this function
def send_confirm_account_email(user, request, cur_app):
    # do not send email in test environment
    if not cur_app.testing:
        # get necessary information out of user object
        name = user.name or "ASReview user"
        # get url of front-end
        root_url = request.headers.get("Origin")
        # create url that will be used in the email
        url = f"{root_url}/confirm_account?user_id={user.id}&token={user.token}"
        # create a mailer
        mailer = Mail(cur_app)
        # open templates as string and render
        root_path = Path(cur_app.root_path)
        with open(
            root_path / "templates" / "emails" / "confirm_account.html", "r"
        ) as f:
            html_text = render_template_string(f.read(), name=name, url=url)
        with open(root_path / "templates" / "emails" / "confirm_account.txt", "r") as f:
            txt_text = render_template_string(f.read(), name=name, url=url)
        # create message
        msg = Message(
            "ASReview: please confirm your account",
            recipients=[user.email]
        )
        msg.body = txt_text
        msg.html = html_text
        return mailer.send(msg)


# ------------------
#      ROUTES
# ------------------


@bp.route("/signin", methods=["POST"])
def signin():
    email = request.form.get("email").strip()
    password = request.form.get("password", "")

    # get the user
    user = User.query.filter(
        or_(User.identifier == email, User.email == email)
    ).one_or_none()

    if not user:
        # user does not exsist
        result = (404, {"message": f"User account {email} does not exist."})
    elif not user.confirmed:
        # account is not confirmed
        result = (404, {"message": f"User account {email} is not confirmed."})
    else:
        # user exists and is confirmed: verify password
        if user.verify_password(password):
            if perform_login_user(user):
                result = (
                    200,
                    {
                        "logged_in": True,
                        "name": user.get_name(),
                        "id": user.id,
                        "message": f"User {user.identifier} is logged in."
                    },
                )
            else:
                result = (
                    404,
                    {"message": "Unable to login user with verified password."}
                )
        else:
            # password is wrong
            if user.origin == "asreview":
                # if this is an asreview user
                result = (404, {"message": f"Incorrect password for user {email}."})
            else:
                # this must be an OAuth user trying to get in with
                # a password
                service = user.origin.capitalize()
                result = (404, {"message": f"Please login with the {service} service."})

    status, message = result
    response = jsonify(message)
    return response, status


@bp.route("/signup", methods=["POST"])
def signup():
    # Can we create accounts?
    if current_app.config.get("ALLOW_ACCOUNT_CREATION", False):
        email = request.form.get("email", "").strip()
        name = request.form.get("name", "").strip()
        affiliation = request.form.get("affiliation", "").strip()
        password = request.form.get("password")
        public = bool(int(request.form.get("public", "1")))

        # check if email already exists
        user = User.query.filter(
            or_(User.identifier == email, User.email == email)
        ).one_or_none()
        # return error if user doesn't exist
        if isinstance(user, User):
            result = (403, f'User with email "{email}" already exists.')
        else:
            try:
                identifier = email
                origin = "asreview"
                # are we going to verify the email?
                email_verification = bool(
                    current_app.config.get("EMAIL_VERIFICATION", False)
                )
                # set confirmed to False if we 'do' verification. Note
                # that this route only creates 'asreview' accounts
                confirmed = not email_verification
                # create the User account
                user = User(
                    identifier=identifier,
                    origin=origin,
                    email=email,
                    name=name,
                    affiliation=affiliation,
                    password=password,
                    confirmed=confirmed,
                    public=public,
                )
                # if this is an un-confirmed account, set token
                if not confirmed:
                    # set token data
                    user = user.set_token_data(
                        current_app.config["SECRET_KEY"],
                        current_app.config["SECURITY_PASSWORD_SALT"],
                    )
                # store user
                DB.session.add(user)
                DB.session.commit()
                # at this stage, if all went well, the User account is
                # stored in the database, send the verification email
                # if applicable
                if email_verification:
                    # send email
                    send_confirm_account_email(user, request, current_app)
                    # result
                    result = (
                        201,
                        f"An email has been sent to {user.email} to verify "
                        + "your account. Please follow instructions.",
                    )
                else:
                    # result is a 201 with message
                    result = (201, f'User "{identifier}" created.')
            except ValueError as e:
                result = (400, f"Unable to create your account! Reason: {str(e)}")
            except IntegrityError as e:
                DB.session.rollback()
                result = (403, f"Unable to create your account! Reason: {str(e)}")
            except SQLAlchemyError as e:
                DB.session.rollback()
                result = (403, f"Unable to create your account! Reason: {str(e)}")
    else:
        result = (400, "The app is not configured to create accounts")

    (status, message) = result
    response = jsonify({"message": message})
    return response, status


@bp.route("/confirm_account", methods=["POST"])
def confirm_account():
    """Confirms account with email verification"""

    if current_app.config.get("EMAIL_VERIFICATION", False):
        # find user by token and user id
        user_id = request.form.get("user_id", 0)
        token = request.form.get("token", "")

        user = User.query.filter(
            and_(User.id == user_id, User.token == token)
        ).one_or_none()

        if not user:
            result = (404, "No user account / correct token found.")
        elif not user.token_valid(token, max_hours=24):
            message = (
                "Can not confirm account, token has expired. "
                + 'Use "forgot password" to obtain a new one.'
            )
            result = (403, message)
        else:
            user = user.confirm_user()
            try:
                DB.session.commit()
                result = (200, f"User {user.identifier} confirmed.")
            except SQLAlchemyError as e:
                DB.session.rollback()
                result = (
                    403,
                    f"Unable to to confirm user {user.identifier}! Reason: {str(e)}"
                )
    else:
        result = (400, "The app is not configured to verify accounts.")

    status, message = result
    response = jsonify({"message": message})
    return response, status


@bp.route("/get_profile", methods=["GET"])
@login_required
def get_profile():
    user = User.query.filter(User.id == current_user.id).one_or_none()
    if user:
        result = (
            200,
            {
                "identifier": user.identifier,
                "email": user.email,
                "origin": user.origin,
                "name": user.name,
                "affiliation": user.affiliation,
                "public": user.public,
            },
        )
    else:
        result = (404, "No user found.")

    status, message = result
    response = jsonify({"message": message})
    return response, status


@bp.route("/forgot_password", methods=["POST"])
def forgot_password():
    if _has_email_configuration(current_app):
        # get email address from request
        email_address = request.form.get("email", "").strip()

        # check if email already exists
        user = User.query.filter(
            or_(User.identifier == email_address, User.email == email_address)
        ).one_or_none()

        if not user:
            result = (404, f'User with email "{email_address}" not found.')
        elif user.origin != "asreview":
            result = (404, f"Your account has been created with {user.origin}.")
        else:
            # set a token
            user = user.set_token_data(
                current_app.config["SECRET_KEY"],
                current_app.config["SECURITY_PASSWORD_SALT"],
            )
            try:
                # store data
                DB.session.commit()
                # send email
                send_forgot_password_email(user, request, current_app)
                # result
                result = (200, f"An email has been sent to {email_address}")

            except SQLAlchemyError as e:
                DB.session.rollback()
                result = (403, f"Unable to to confirm user! Reason: {str(e)}")
    else:
        result = (404, "Forgot-password feature is not used in this app.")

    status, message = result
    response = jsonify({"message": message})
    return response, status


@bp.route("/reset_password", methods=["POST"])
def reset_password():
    """Resests password of user"""
    if _has_email_configuration(current_app):

        new_password = request.form.get("password", "").strip()
        token = request.form.get("token", "").strip()
        user_id = request.form.get("user_id", "0").strip()
        user = User.query.filter(User.id == user_id).one_or_none()

        if not user:
            result = (
                404,
                "User not found, try restarting the forgot-password procedure."
            )
        elif not user.token_valid(token, max_hours=24):
            result = (
                404,
                "Token is invalid or too old, restart the forgot-password procedure."
            )
        else:
            try:
                user = user.reset_password(new_password)
                DB.session.commit()
                result = (200, "Password updated.")
            except ValueError as e:
                DB.session.rollback()
                result = (500, f"Unable to reset your password! Reason: {str(e)}")
            except SQLAlchemyError as e:
                DB.session.rollback()
                result = (500, f"Unable to reset your password! Reason: {str(e)}")
    else:
        result = (404, "Reset-password feature is not used in this app.")

    status, message = result
    response = jsonify({"message": message})
    return response, status


@bp.route("/update_profile", methods=["POST"])
@login_required
def update_profile():
    """Update user profile"""
    user = User.query.filter(User.id == current_user.id).one_or_none()
    if user:
        email = request.form.get("email", "").strip()
        name = request.form.get("name", "").strip()
        affiliation = request.form.get("affiliation", "").strip()
        old_password = request.form.get("old_password", None)
        new_password = request.form.get("new_password", None)
        public = bool(int(request.form.get("public", "1")))

        try:
            user = user.update_profile(email, name, affiliation,
                                       old_password, new_password, public)
            DB.session.commit()
            result = (200, "User profile updated.")
        except ValueError as e:
            result = (400, f"Unable to update your profile! Reason: {str(e)}")
        except IntegrityError as e:
            DB.session.rollback()
            result = (500, f"Unable to update your profile! Reason: {str(e)}")
        except SQLAlchemyError as e:
            DB.session.rollback()
            result = (500, f"Unable to update your profile! Reason: {str(e)}")

    else:
        result = (404, "No user found")

    status, message = result
    response = jsonify({"message": message})
    return response, status


@bp.route("/refresh", methods=["GET"])
def refresh():
    if current_user and isinstance(current_user, User):
        logged_in = current_user.is_authenticated
        name = current_user.get_name()
        id = current_user.id
    else:
        logged_in = False
        name = ""
        id = None

    result = {"logged_in": logged_in, "name": name, "id": id}

    response = jsonify(result)
    return response, 200


@bp.route("/signout", methods=["DELETE"])
@login_required
def signout():
    if current_user:
        identifier = current_user.identifier
        logout_user()
        result = (200, f"User with identifier {identifier} has been signed out.")
    else:
        result = (404, "No user found, no one can be signed out.")

    status, message = result
    response = jsonify({"message": message})
    return response, status


@bp.route("/oauth_callback", methods=["POST"])
def oauth_callback():
    # get parameters
    code = request.form.get("code", "").strip()
    provider = request.form.get("provider", "").strip()
    redirect_uri = request.form.get("redirect_uri", "").strip()

    # assuming we have this provider
    oauth_handler = current_app.config.get("OAUTH", False)
    if (
        isinstance(oauth_handler, OAuthHandler)
        and provider in oauth_handler.providers()
    ):
        # get user credentials for this user
        (identifier, email, name) = oauth_handler.get_user_credentials(
            provider, code, redirect_uri
        )
        # make sure identifier is a string
        identifier = str(identifier)
        # try to find this user
        user = User.query.filter(User.identifier == identifier).one_or_none()
        # flag for response (I'd like to communicate if this user was created)
        created_account = False
        # if not create user
        if user is None:
            try:
                origin = provider
                confirmed = True
                public = True
                user = User(
                    identifier=identifier,
                    origin=origin,
                    email=email,
                    name=name,
                    confirmed=confirmed,
                    public=public,
                )
                DB.session.add(user)
                DB.session.commit()
                created_account = True
            except IntegrityError:
                DB.session.rollback()
                message = "OAuth: integrity error, verify if you " + \
                    "already have created an account!"
                # return this immediately
                return jsonify({"message": message}), 409
            except SQLAlchemyError:
                DB.session.rollback()
                message = "OAuth: unable to create your account!"
                # return this immediately
                return jsonify({"message": message}), 409

        # log in the existing/created user immediately
        logged_in = perform_login_user(user)
        result = (
            200,
            {
                "account_created": created_account,
                "logged_in": logged_in,
                "name": user.get_name(),
                "id": user.id,
            },
        )
    else:
        result = (400, {"message": f"OAuth provider {provider} could not be found"})

    status, message = result
    response = jsonify(message)
    return response, status



# Copyright 2019-2022 The ASReview Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

__all__ = ["LSTMPoolClassifier"]

import logging

try:
    import tensorflow as tf
    from tensorflow.keras.constraints import MaxNorm
    from tensorflow.keras.layers import LSTM
    from tensorflow.keras.layers import Dense
    from tensorflow.keras.layers import Embedding
    from tensorflow.keras.layers import Flatten
    from tensorflow.keras.layers import MaxPooling1D
    from tensorflow.keras.models import Sequential
except ImportError:
    TF_AVAILABLE = False
else:
    TF_AVAILABLE = True
    try:
        tf.logging.set_verbosity(tf.logging.ERROR)
    except AttributeError:
        logging.getLogger("tensorflow").setLevel(logging.ERROR)

import numpy as np
import scipy

from asreview.models.classifiers.base import BaseTrainClassifier
from asreview.models.classifiers.lstm_base import _get_optimizer
from asreview.models.classifiers.utils import _set_class_weight


def _check_tensorflow():
    if not TF_AVAILABLE:
        raise ImportError("Install tensorflow package to use" " LSTM-pool.")


class LSTMPoolClassifier(BaseTrainClassifier):
    """
    LSTM-pool classifier (``lstm-pool``).

    LSTM model that consists of an embedding layer, LSTM layer with many
    outputs, max pooling layer, and a single sigmoid output node. Use the
    :class:`asreview.models.feature_extraction.EmbeddingLSTM` feature extraction
    method. Currently not so well optimized and slow.

    .. note::

        This model requires ``tensorflow`` to be installed. Use ``pip install
        asreview[tensorflow]`` or install all optional ASReview dependencies
        with ``pip install asreview[all]``

    Arguments
    ---------
    embedding_matrix: numpy.ndarray
        Embedding matrix to use with LSTM model.
    backwards: bool
        Whether to have a forward or backward LSTM.
    dropout: float
        Value in [0, 1.0) that gives the dropout and recurrent
        dropout rate for the LSTM model.
    optimizer: str
        Optimizer to use.
    lstm_out_width: int
        Output width of the LSTM.
    lstm_pool_size: int
        Size of the pool, must be a divisor of max_sequence_length.
    learn_rate: float
        Learn rate multiplier of default learning rate.
    verbose: int
        Verbosity.
    batch_size: int
        Size of the batch size for the LSTM model.
    epochs: int
        Number of epochs to train the LSTM model.
    shuffle: bool
        Whether to shuffle the data before starting to train.
    class_weight: float
        Class weight for the included papers.
    """

    name = "lstm-pool"
    label = "LSTM with a max pooling layer"

    def __init__(
        self,
        embedding_matrix=None,
        backwards=True,
        dropout=0.4,
        optimizer="rmsprop",
        lstm_out_width=20,
        lstm_pool_size=128,
        learn_rate=1.0,
        verbose=0,
        batch_size=32,
        epochs=35,
        shuffle=False,
        class_weight=30.0,
    ):
        """Initialize the LSTM pool model."""
        super(LSTMPoolClassifier, self).__init__()
        self.embedding_matrix = embedding_matrix
        self.backwards = backwards
        self.dropout = dropout
        self.optimizer = optimizer
        self.lstm_out_width = lstm_out_width
        self.learn_rate = learn_rate
        self.verbose = verbose
        self.batch_size = batch_size
        self.epochs = epochs
        self.shuffle = shuffle
        self.class_weight = _set_class_weight(class_weight)
        self.lstm_pool_size = lstm_pool_size
        self._model = None
        self.sequence_length = None

    def fit(self, X, y):
        # check is tensorflow is available
        _check_tensorflow()

        if scipy.sparse.isspmatrix(X):
            X = X.toarray()

        sequence_length = X.shape[1]
        if self._model is None or sequence_length != self.sequence_length:
            self.sequence_length = sequence_length
            self._model = _create_lstm_pool_model(
                embedding_matrix=self.embedding_matrix,
                backwards=self.backwards,
                dropout=self.dropout,
                optimizer=self.optimizer,
                max_sequence_length=sequence_length,
                lstm_out_width=self.lstm_out_width,
                learn_rate=self.learn_rate,
                verbose=self.verbose,
            )

        self._model.fit(
            X,
            y,
            batch_size=self.batch_size,
            epochs=self.epochs,
            shuffle=self.shuffle,
            class_weight=self.class_weight,
            verbose=self.verbose,
        )

    def predict_proba(self, X):
        if scipy.sparse.isspmatrix(X):
            X = X.toarray()
        pos_pred = self._model.predict(X, verbose=self.verbose)
        neg_pred = 1 - pos_pred
        return np.hstack([neg_pred, pos_pred])

    def full_hyper_space(self):
        from hyperopt import hp

        hyper_choices = {}
        hyper_space = {
            "mdl_dropout": hp.uniform("mdl_dropout", 0, 0.9),
            "mdl_lstm_out_width": hp.quniform("mdl_lstm_out_width", 1, 50, 1),
            "mdl_dense_width": hp.quniform("mdl_dense_width", 1, 200, 1),
            "mdl_learn_rate_mult": hp.lognormal("mdl_learn_rate_mult", 0, 1),
        }
        return hyper_space, hyper_choices

    @property
    def default_param(self):
        defaults = super(LSTMPoolClassifier, self).default_param
        defaults.pop("embedding_matrix")
        return defaults


def _create_lstm_pool_model(
    embedding_matrix,
    backwards=True,
    dropout=0.4,
    optimizer="rmsprop",
    max_sequence_length=1000,
    lstm_out_width=20,
    lstm_pool_size=100,
    learn_rate=1.0,
    verbose=1,
):
    """Return callable lstm model.
    Returns
    -------
    callable:
        A function that return the Keras Sklearn model when
        called.

    """

    # check is tensorflow is available
    _check_tensorflow()

    # The Sklearn API requires a callable as result.
    # https://keras.io/scikit-learn-api/
    model = Sequential()

    # add first embedding layer with pretrained wikipedia weights
    model.add(
        Embedding(
            embedding_matrix.shape[0],
            embedding_matrix.shape[1],
            weights=[embedding_matrix],
            input_length=max_sequence_length,
            trainable=False,
        )
    )

    # add LSTM layer
    model.add(
        LSTM(
            lstm_out_width,
            input_shape=(max_sequence_length,),
            go_backwards=backwards,
            dropout=dropout,
            recurrent_dropout=dropout,
            return_sequences=True,
            kernel_constraint=MaxNorm(),
        )
    )

    model.add(
        MaxPooling1D(
            pool_size=lstm_pool_size,
        )
    )
    model.add(Flatten())

    # Add output layer
    model.add(Dense(1, activation="sigmoid"))

    optimizer_fn = _get_optimizer(optimizer, learn_rate)

    # Compile model
    model.compile(
        loss="binary_crossentropy", optimizer=optimizer_fn, metrics=["acc"]
    )

    if verbose >= 1:
        model.summary()

    return model



# Copyright 2019-2022 The ASReview Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

__all__ = [
    "COLUMN_DEFINITIONS",
    "DEFAULT_BALANCE_STRATEGY",
    "DEFAULT_FEATURE_EXTRACTION",
    "DEFAULT_MODEL",
    "DEFAULT_N_INSTANCES",
    "DEFAULT_N_PRIOR_EXCLUDED",
    "DEFAULT_N_PRIOR_INCLUDED",
    "DEFAULT_QUERY_STRATEGY",
    "KERAS_MODELS",
    "LABEL_NA",
    "PROJECT_MODE_EXPLORE",
    "PROJECT_MODE_ORACLE",
    "PROJECT_MODE_SIMULATE",
    "PROJECT_MODES",
    "SCHEMA",
]

LABEL_NA = -1

KERAS_MODELS = ["lstm_base", "lstm_pool"]

# project types
PROJECT_MODE_ORACLE = "oracle"
PROJECT_MODE_EXPLORE = "explore"
PROJECT_MODE_SIMULATE = "simulate"
PROJECT_MODES = [PROJECT_MODE_ORACLE, PROJECT_MODE_EXPLORE, PROJECT_MODE_SIMULATE]

# CLI defaults
DEFAULT_MODEL = "nb"
DEFAULT_QUERY_STRATEGY = "max"
DEFAULT_BALANCE_STRATEGY = "double"
DEFAULT_FEATURE_EXTRACTION = "tfidf"
DEFAULT_N_INSTANCES = 1
DEFAULT_N_PRIOR_INCLUDED = 1
DEFAULT_N_PRIOR_EXCLUDED = 1

COLUMN_DEFINITIONS = {
    # included
    "included": [
        "final_included",
        "label",
        "label_included",
        "included_label",
        "included_final",
        "included",
        "included_flag",
        "include",
    ],
    # abstract included (pending deprecation)
    "abstract_included": [
        "abstract_included",
        "included_abstract",
        "included_after_abstract",
        "label_abstract_screening",
    ],
    "title": ["title", "primary_title"],
    "authors": ["authors", "author names", "first_authors"],
    "abstract": ["abstract", "abstract note", "notes_abstract"],
    "notes": ["notes"],
    "keywords": ["keywords"],
    "doi": ["doi"],
}

# the schema describes the content of the ASReview project file.
SCHEMA = {
    "$schema": "http://json-schema.org/draft-07/schema",
    "$id": "http://example.com/example.json",
    "type": "object",
    "title": "The ASReview project file root schema",
    "description": "The root schema comprises the entire project.json file in the ASReview project file.",
    "default": {},
    "examples": [
        {
            "version": "1.0",
            "id": "example",
            "mode": "oracle",
            "name": "example",
            "description": "",
            "authors": "",
            "created_at_unix": 1648205610,
            "datetimeCreated": "2022-03-25 11:53:30.510461",
            "reviews": [
                {
                    "id": "4793de70a8d44eb4baa68bac2853c91a",
                    "start_time": "2022-03-25 11:55:50.551360",
                    "status": "review",
                    "end_time": "2022-03-26 10:31:52.441360",
                }
            ],
            "feature_matrices": [
                {"id": "tfidf", "filename": "tfidf_feature_matrix.npz"}
            ],
            "dataset_path": "example.ris",
        }
    ],
    "required": ["version", "id", "mode", "name"],
    "properties": {
        "version": {
            "$id": "#/properties/version",
            "type": "string",
            "title": "The version schema",
            "description": "The version of ASReview on initiation of the project.",
            "default": "",
            "examples": ["1.0"],
        },
        "id": {
            "$id": "#/properties/id",
            "type": "string",
            "title": "The id schema",
            "description": "The unique identifier of the project.",
            "default": "",
            "examples": ["example"],
        },
        "mode": {
            "$id": "#/properties/mode",
            "type": "string",
            "title": "The mode schema",
            "description": "The mode of the project. One of oracle, explore, or simulate.",
            "default": "",
            "enum": PROJECT_MODES,
            "examples": ["oracle"],
        },
        "name": {
            "$id": "#/properties/name",
            "type": ["string", "null"],
            "title": "The name schema",
            "description": "The name of the project.",
            "default": "",
            "examples": ["example"],
        },
        "description": {
            "$id": "#/properties/description",
            "type": ["string", "null"],
            "title": "The description schema",
            "description": "The description of the project.",
            "default": "",
            "examples": [""],
        },
        "authors": {
            "$id": "#/properties/authors",
            "type": ["string", "null"],
            "title": "The authors schema",
            "description": "The authors of the project.",
            "default": "",
            "examples": [""],
        },
        "created_at_unix": {
            "$id": "#/properties/created_at_unix",
            "type": ["integer", "null"],
            "title": "The created_at_unix schema",
            "description": "An explanation about the purpose of this instance.",
            "default": 0,
            "examples": [1648205610],
        },
        "datetimeCreated": {
            "$id": "#/properties/datetimeCreated",
            "type": ["string", "null"],
            "title": "The datetimeCreated schema",
            "description": "The date and time of the project creation.",
            "default": "",
            "examples": ["2022-03-25 11:53:30.510461"],
        },
        "reviews": {
            "$id": "#/properties/reviews",
            "type": "array",
            "title": "The reviews schema",
            "description": "The list of reviews in the project. Multiple reviews per project are possible, however this is limited to 1 at the moment.",
            "default": [],
            "examples": [
                [
                    {
                        "id": "4793de70a8d44eb4baa68bac2853c91a",
                        "start_time": "2022-03-25 11:55:50.551360",
                        "status": "review",
                    }
                ]
            ],
            "additionalItems": True,
            "items": {
                "$id": "#/properties/reviews/items",
                "anyOf": [
                    {
                        "$id": "#/properties/reviews/items/anyOf/0",
                        "type": "object",
                        "title": "The first anyOf schema",
                        "description": "An explanation about the purpose of this instance.",
                        "default": {},
                        "examples": [
                            {
                                "id": "4793de70a8d44eb4baa68bac2853c91a",
                                "start_time": "2022-03-25 11:55:50.551360",
                                "status": "review",
                            }
                        ],
                        "required": ["id", "start_time", "status"],
                        "properties": {
                            "id": {
                                "$id": "#/properties/reviews/items/anyOf/0/properties/id",
                                "type": "string",
                                "title": "The id of the review.",
                                "description": "A unique UUID4 identifier of the review.",
                                "default": "",
                                "examples": ["4793de70a8d44eb4baa68bac2853c91a"],
                            },
                            "start_time": {
                                "$id": "#/properties/reviews/items/anyOf/0/properties/start_time",
                                "type": "string",
                                "title": "The start_time of the review.",
                                "description": "The start date and time of the review.",
                                "default": "",
                                "examples": ["2022-03-25 11:55:50.551360"],
                            },
                            "end_time": {
                                "$id": "#/properties/reviews/items/anyOf/0/properties/start_time",
                                "type": "string",
                                "title": "The end_time of the review.",
                                "description": "The end date and time of the review.",
                                "default": "",
                                "examples": ["2022-03-26 10:31:52.441360"],
                            },
                            "status": {
                                "$id": "#/properties/reviews/items/anyOf/0/properties/status",
                                "type": ["string", "null"],
                                "title": "The status of the review.",
                                "description": "The status of the review. Options are setup, review, finished.",
                                "enum": ["setup", "review", "finished"],
                                "default": "setup",
                                "examples": ["review"],
                            },
                        },
                        "additionalProperties": True,
                    }
                ],
            },
        },
        "feature_matrices": {
            "$id": "#/properties/feature_matrices",
            "type": "array",
            "title": "The feature_matrices schema",
            "description": "Information about the feature matrices.",
            "default": [],
            "examples": [[{"id": "tfidf", "filename": "tfidf_feature_matrix.npz"}]],
            "additionalItems": True,
            "items": {
                "$id": "#/properties/feature_matrices/items",
                "anyOf": [
                    {
                        "$id": "#/properties/feature_matrices/items/anyOf/0",
                        "type": "object",
                        "title": "The first anyOf schema",
                        "description": "Information about a feature matrix.",
                        "default": {},
                        "examples": [
                            {"id": "tfidf", "filename": "tfidf_feature_matrix.npz"}
                        ],
                        "required": ["id", "filename"],
                        "properties": {
                            "id": {
                                "$id": "#/properties/feature_matrices/items/anyOf/0/properties/id",
                                "type": "string",
                                "title": "The id schema",
                                "description": "A unique id of the feature matrix.",
                                "default": "",
                                "examples": ["tfidf"],
                            },
                            "filename": {
                                "$id": "#/properties/feature_matrices/items/anyOf/0/properties/filename",
                                "type": "string",
                                "title": "The filename schema",
                                "description": "The name of the file with the feature matrix. Usually a sparse matrix.",
                                "default": "",
                                "examples": ["tfidf_feature_matrix.npz"],
                            },
                        },
                        "additionalProperties": True,
                    }
                ],
            },
        },
        "dataset_path": {
            "$id": "#/properties/dataset_path",
            "type": ["string", "null"],
            "title": "The dataset_path schema",
            "description": "Name of the dataset file.",
            "default": "",
            "examples": ["example.ris"],
        },
    },
    "additionalProperties": True,
}
